@@ -7,52 +7,60 @@ Protocol objects representing different implementations of the same classes.
 from typing import Any, Callable, Generator, Mapping
 from typing import List, Optional, Sequence, Tuple, TypeVar, Union
 from typing import TYPE_CHECKING
+from typing_extensions import TypeAlias
 
 from . import pq
 from ._enums import PyFormat as PyFormat
-from ._compat import Protocol
+from ._compat import Protocol, LiteralString
 
 if TYPE_CHECKING:
-    from .sql import Composable
+    from . import sql
     from .rows import Row, RowMaker
     from .pq.abc import PGresult
     from .waiting import Wait, Ready
     from .connection import BaseConnection
     from ._adapters_map import AdaptersMap
 
+NoneType: type = type(None)
+
 # An object implementing the buffer protocol
-Buffer = Union[bytes, bytearray, memoryview]
+Buffer: TypeAlias = Union[bytes, bytearray, memoryview]
 
-Query = Union[str, bytes, "Composable"]
-Params = Union[Sequence[Any], Mapping[str, Any]]
+Query: TypeAlias = Union[LiteralString, bytes, "sql.SQL", "sql.Composed"]
+Params: TypeAlias = Union[Sequence[Any], Mapping[str, Any]]
 ConnectionType = TypeVar("ConnectionType", bound="BaseConnection[Any]")
-
-# TODO: make it recursive when mypy will support it
-# DumperKey = Union[type, Tuple[Union[type, "DumperKey"]]]
-DumperKey = Union[type, Tuple[type, ...]]
+PipelineCommand: TypeAlias = Callable[[], None]
+DumperKey: TypeAlias = Union[type, Tuple["DumperKey", ...]]
 
 # Waiting protocol types
 
 RV = TypeVar("RV")
 
-PQGenConn = Generator[Tuple[int, "Wait"], "Ready", RV]
+PQGenConn: TypeAlias = Generator[Tuple[int, "Wait"], "Ready", RV]
 """Generator for processes where the connection file number can change.
 
 This can happen in connection and reset, but not in normal querying.
 """
 
-PQGen = Generator["Wait", "Ready", RV]
+PQGen: TypeAlias = Generator["Wait", "Ready", RV]
 """Generator for processes where the connection file number won't change.
-
-The first item generated is the file descriptor; following items are be the
-Wait states.
 """
 
 
+class WaitFunc(Protocol):
+    """
+    Wait on the connection which generated `PQgen` and return its final result.
+    """
+
+    def __call__(
+        self, gen: PQGen[RV], fileno: int, timeout: Optional[float] = None
+    ) -> RV: ...
+
+
 # Adaptation types
 
-DumpFunc = Callable[[Any], bytes]
-LoadFunc = Callable[[bytes], Any]
+DumpFunc: TypeAlias = Callable[[Any], Buffer]
+LoadFunc: TypeAlias = Callable[[Buffer], Any]
 
 
 class AdaptContext(Protocol):
@@ -83,7 +91,7 @@ class AdaptContext(Protocol):
 
 class Dumper(Protocol):
     """
-    Convert Python objects of type *cls* to PostgreSQL representation.
+    Convert Python objects of type `!cls` to PostgreSQL representation.
     """
 
     format: pq.Format
@@ -97,25 +105,24 @@ class Dumper(Protocol):
     oid: int
     """The oid to pass to the server, if known; 0 otherwise (class attribute)."""
 
-    def __init__(self, cls: type, context: Optional[AdaptContext] = None):
-        ...
+    def __init__(self, cls: type, context: Optional[AdaptContext] = None): ...
 
     def dump(self, obj: Any) -> Buffer:
-        """Convert the object *obj* to PostgreSQL representation.
+        """Convert the object `!obj` to PostgreSQL representation.
 
         :param obj: the object to convert.
         """
         ...
 
     def quote(self, obj: Any) -> Buffer:
-        """Convert the object *obj* to escaped representation.
+        """Convert the object `!obj` to escaped representation.
 
         :param obj: the object to convert.
         """
         ...
 
     def get_key(self, obj: Any, format: PyFormat) -> DumperKey:
-        """Return an alternative key to upgrade the dumper to represent *obj*.
+        """Return an alternative key to upgrade the dumper to represent `!obj`.
 
         :param obj: The object to convert
         :param format: The format to convert to
@@ -140,9 +147,9 @@ class Dumper(Protocol):
           size of int?...)
 
         In these cases, a dumper can implement `!get_key()` and return a new
-        class, or sequence of classes, that can be used to indentify the same
+        class, or sequence of classes, that can be used to identify the same
         dumper again. If the mechanism is not needed, the method should return
-        the same *cls* object passed in the constructor.
+        the same `!cls` object passed in the constructor.
 
         If a dumper implements `get_key()` it should also implement
         `upgrade()`.
@@ -151,13 +158,13 @@ class Dumper(Protocol):
         ...
 
     def upgrade(self, obj: Any, format: PyFormat) -> "Dumper":
-        """Return a new dumper to manage *obj*.
+        """Return a new dumper to manage `!obj`.
 
         :param obj: The object to convert
         :param format: The format to convert to
 
         Once `Transformer.get_dumper()` has been notified by `get_key()` that
-        this Dumper class cannot handle *obj* itself, it will invoke
+        this Dumper class cannot handle `!obj` itself, it will invoke
         `!upgrade()`, which should return a new `Dumper` instance, which will
         be reused for every objects for which `!get_key()` returns the same
         result.
@@ -167,7 +174,7 @@ class Dumper(Protocol):
 
 class Loader(Protocol):
     """
-    Convert PostgreSQL values with type OID *oid* to Python objects.
+    Convert PostgreSQL values with type OID `!oid` to Python objects.
     """
 
     format: pq.Format
@@ -178,8 +185,7 @@ class Loader(Protocol):
     This is a class attribute.
     """
 
-    def __init__(self, oid: int, context: Optional[AdaptContext] = None):
-        ...
+    def __init__(self, oid: int, context: Optional[AdaptContext] = None): ...
 
     def load(self, data: Buffer) -> Any:
         """
@@ -191,24 +197,25 @@ class Loader(Protocol):
 
 
 class Transformer(Protocol):
-
     types: Optional[Tuple[int, ...]]
     formats: Optional[List[pq.Format]]
 
-    def __init__(self, context: Optional[AdaptContext] = None):
-        ...
+    def __init__(self, context: Optional[AdaptContext] = None): ...
+
+    @classmethod
+    def from_context(cls, context: Optional[AdaptContext]) -> "Transformer": ...
 
     @property
-    def connection(self) -> Optional["BaseConnection[Any]"]:
-        ...
+    def connection(self) -> Optional["BaseConnection[Any]"]: ...
 
     @property
-    def adapters(self) -> "AdaptersMap":
-        ...
+    def encoding(self) -> str: ...
 
     @property
-    def pgresult(self) -> Optional["PGresult"]:
-        ...
+    def adapters(self) -> "AdaptersMap": ...
+
+    @property
+    def pgresult(self) -> Optional["PGresult"]: ...
 
     def set_pgresult(
         self,
@@ -216,39 +223,26 @@ class Transformer(Protocol):
         *,
         set_loaders: bool = True,
         format: Optional[pq.Format] = None
-    ) -> None:
-        ...
+    ) -> None: ...
 
-    def set_dumper_types(
-        self, types: Sequence[int], format: pq.Format
-    ) -> None:
-        ...
+    def set_dumper_types(self, types: Sequence[int], format: pq.Format) -> None: ...
 
-    def set_loader_types(
-        self, types: Sequence[int], format: pq.Format
-    ) -> None:
-        ...
+    def set_loader_types(self, types: Sequence[int], format: pq.Format) -> None: ...
 
     def dump_sequence(
         self, params: Sequence[Any], formats: Sequence[PyFormat]
-    ) -> Sequence[Optional[Buffer]]:
-        ...
+    ) -> Sequence[Optional[Buffer]]: ...
 
-    def get_dumper(self, obj: Any, format: PyFormat) -> Dumper:
-        ...
+    def as_literal(self, obj: Any) -> bytes: ...
+
+    def get_dumper(self, obj: Any, format: PyFormat) -> Dumper: ...
 
     def load_rows(
         self, row0: int, row1: int, make_row: "RowMaker[Row]"
-    ) -> List["Row"]:
-        ...
+    ) -> List["Row"]: ...
 
-    def load_row(self, row: int, make_row: "RowMaker[Row]") -> Optional["Row"]:
-        ...
+    def load_row(self, row: int, make_row: "RowMaker[Row]") -> Optional["Row"]: ...
 
-    def load_sequence(
-        self, record: Sequence[Optional[bytes]]
-    ) -> Tuple[Any, ...]:
-        ...
+    def load_sequence(self, record: Sequence[Optional[Buffer]]) -> Tuple[Any, ...]: ...
 
-    def get_loader(self, oid: int, format: pq.Format) -> Loader:
-        ...
+    def get_loader(self, oid: int, format: pq.Format) -> Loader: ...
