@@ -5,7 +5,7 @@ Entry point into the adaptation system.
 # Copyright (C) 2020 The Psycopg Team
 
 from abc import ABC, abstractmethod
-from typing import Any, Optional, Type, Tuple, Union, TYPE_CHECKING
+from typing import Any, Optional, Type, TYPE_CHECKING
 
 from . import pq, abc
 from . import _adapters_map
@@ -23,7 +23,7 @@ ORD_BS = ord("\\")
 
 class Dumper(abc.Dumper, ABC):
     """
-    Convert Python object of the type *cls* to PostgreSQL representation.
+    Convert Python object of the type `!cls` to PostgreSQL representation.
     """
 
     oid: int = 0
@@ -45,8 +45,7 @@ class Dumper(abc.Dumper, ABC):
         )
 
     @abstractmethod
-    def dump(self, obj: Any) -> Buffer:
-        ...
+    def dump(self, obj: Any) -> Buffer: ...
 
     def quote(self, obj: Any) -> Buffer:
         """
@@ -91,12 +90,12 @@ class Dumper(abc.Dumper, ABC):
             rv = rv.replace(b"\\", b"\\\\")
         return rv
 
-    def get_key(self, obj: Any, format: PyFormat) -> Union[type, Tuple[type, ...]]:
+    def get_key(self, obj: Any, format: PyFormat) -> abc.DumperKey:
         """
         Implementation of the `~psycopg.abc.Dumper.get_key()` member of the
         `~psycopg.abc.Dumper` protocol. Look at its definition for details.
 
-        This implementation returns the *cls* passed in the constructor.
+        This implementation returns the `!cls` passed in the constructor.
         Subclasses needing to specialise the PostgreSQL type according to the
         *value* of the object dumped (not only according to to its type)
         should override this class.
@@ -109,7 +108,7 @@ class Dumper(abc.Dumper, ABC):
         Implementation of the `~psycopg.abc.Dumper.upgrade()` member of the
         `~psycopg.abc.Dumper` protocol. Look at its definition for details.
 
-        This implementation just returns *self*. If a subclass implements
+        This implementation just returns `!self`. If a subclass implements
         `get_key()` it should probably override `!upgrade()` too.
         """
         return self
@@ -117,7 +116,7 @@ class Dumper(abc.Dumper, ABC):
 
 class Loader(abc.Loader, ABC):
     """
-    Convert PostgreSQL values with type OID *oid* to Python objects.
+    Convert PostgreSQL values with type OID `!oid` to Python objects.
     """
 
     format: pq.Format = pq.Format.TEXT
@@ -151,7 +150,7 @@ class RecursiveDumper(Dumper):
 
     def __init__(self, cls: type, context: Optional[abc.AdaptContext] = None):
         super().__init__(cls, context)
-        self._tx = Transformer(context)
+        self._tx = Transformer.from_context(context)
 
 
 class RecursiveLoader(Loader):
@@ -159,4 +158,4 @@ class RecursiveLoader(Loader):
 
     def __init__(self, oid: int, context: Optional[abc.AdaptContext] = None):
         super().__init__(oid, context)
-        self._tx = Transformer(context)
+        self._tx = Transformer.from_context(context)
