@@ -7,43 +7,63 @@ psycopg connection objects
 import logging
 import threading
 from types import TracebackType
-from typing import Any, Callable, cast, Dict, Generator, Generic, Iterator
-from typing import List, NamedTuple, Optional, Type, TypeVar, Union
+from typing import Any, Callable, cast, Generator, Generic, Iterator
+from typing import List, NamedTuple, Optional, Type, TypeVar, Tuple, Union
 from typing import overload, TYPE_CHECKING
 from weakref import ref, ReferenceType
 from warnings import warn
 from functools import partial
 from contextlib import contextmanager
+from typing_extensions import TypeAlias
 
 from . import pq
 from . import errors as e
 from . import waiting
 from . import postgres
-from .pq import ConnStatus, ExecStatus, TransactionStatus, Format
-from .abc import AdaptContext, ConnectionType, Params, Query, RV
+from .abc import AdaptContext, Params, Query, RV
 from .abc import PQGen, PQGenConn
-from .sql import Composable
-from .rows import Row, RowFactory, tuple_row, TupleRow
+from .sql import Composable, SQL
+from ._tpc import Xid
+from .rows import Row, RowFactory, tuple_row, TupleRow, args_row
 from .adapt import AdaptersMap
 from ._enums import IsolationLevel
 from .cursor import Cursor
-from .conninfo import make_conninfo, conninfo_to_dict, ConnectionInfo
-from .generators import connect, execute, notifies
+from ._compat import LiteralString, Self
+from .pq.misc import connection_summary
+from .conninfo import make_conninfo, conninfo_to_dict
+from .conninfo import conninfo_attempts, ConnDict, timeout_from_conninfo
+from ._pipeline import BasePipeline, Pipeline
+from .generators import notifies, connect, execute
 from ._encodings import pgconn_encoding
 from ._preparing import PrepareManager
 from .transaction import Transaction
 from .server_cursor import ServerCursor
+from ._connection_info import ConnectionInfo
 
 if TYPE_CHECKING:
     from .pq.abc import PGconn, PGresult
     from psycopg_pool.base import BasePool
 
-logger = logging.getLogger("psycopg")
 
 # Row Type variable for Cursor (when it needs to be distinguished from the
 # connection's one)
 CursorRow = TypeVar("CursorRow")
 
+TEXT = pq.Format.TEXT
+BINARY = pq.Format.BINARY
+
+OK = pq.ConnStatus.OK
+BAD = pq.ConnStatus.BAD
+
+COMMAND_OK = pq.ExecStatus.COMMAND_OK
+TUPLES_OK = pq.ExecStatus.TUPLES_OK
+FATAL_ERROR = pq.ExecStatus.FATAL_ERROR
+
+IDLE = pq.TransactionStatus.IDLE
+INTRANS = pq.TransactionStatus.INTRANS
+
+logger = logging.getLogger("psycopg")
+
 
 class Notify(NamedTuple):
     """An asynchronous notification received from the database."""
@@ -60,8 +80,8 @@ class Notify(NamedTuple):
 
 Notify.__module__ = "psycopg"
 
-NoticeHandler = Callable[[e.Diagnostic], None]
-NotifyHandler = Callable[[Notify], None]
+NoticeHandler: TypeAlias = Callable[[e.Diagnostic], None]
+NotifyHandler: TypeAlias = Callable[[Notify], None]
 
 
 class BaseConnection(Generic[Row]):
@@ -91,7 +111,10 @@ class BaseConnection(Generic[Row]):
     def __init__(self, pgconn: "PGconn"):
         self.pgconn = pgconn
         self._autocommit = False
-        self._adapters = AdaptersMap(postgres.adapters)
+
+        # None, but set to a copy of the global adapters map as soon as requested.
+        self._adapters: Optional[AdaptersMap] = None
+
         self._notice_handlers: List[NoticeHandler] = []
         self._notify_handlers: List[NotifyHandler] = []
 
@@ -100,6 +123,7 @@ class BaseConnection(Generic[Row]):
 
         self._closed = False  # closed by an explicit close()
         self._prepared: PrepareManager = PrepareManager()
+        self._tpc: Optional[Tuple[Xid, bool]] = None  # xid, prepared
 
         wself = ref(self)
         pgconn.notice_handler = partial(BaseConnection._notice_handler, wself)
@@ -109,6 +133,8 @@ class BaseConnection(Generic[Row]):
         # apart a connection in the pool too (when _pool = None)
         self._pool: Optional["BasePool[Any]"]
 
+        self._pipeline: Optional[BasePipeline] = None
+
         # Time after which the connection should be closed
         self._expire_at: float
 
@@ -138,13 +164,13 @@ class BaseConnection(Generic[Row]):
 
     def __repr__(self) -> str:
         cls = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        info = pq.misc.connection_summary(self.pgconn)
+        info = connection_summary(self.pgconn)
         return f"<{cls} {info} at 0x{id(self):x}>"
 
     @property
     def closed(self) -> bool:
         """`!True` if the connection is closed."""
-        return self.pgconn.status == ConnStatus.BAD
+        return self.pgconn.status == BAD
 
     @property
     def broken(self) -> bool:
@@ -152,9 +178,9 @@ class BaseConnection(Generic[Row]):
         `!True` if the connection was interrupted.
 
         A broken connection is always `closed`, but wasn't closed in a clean
-        way, such as using `close()` or a ``with`` block.
+        way, such as using `close()` or a `!with` block.
         """
-        return self.pgconn.status == ConnStatus.BAD and not self._closed
+        return self.pgconn.status == BAD and not self._closed
 
     @property
     def autocommit(self) -> bool:
@@ -166,9 +192,10 @@ class BaseConnection(Generic[Row]):
         self._set_autocommit(value)
 
     def _set_autocommit(self, value: bool) -> None:
-        # Base implementation, not thread safe.
-        # Subclasses must call it holding a lock
-        self._check_intrans("autocommit")
+        raise NotImplementedError
+
+    def _set_autocommit_gen(self, value: bool) -> PQGen[None]:
+        yield from self._check_intrans_gen("autocommit")
         self._autocommit = bool(value)
 
     @property
@@ -183,9 +210,10 @@ class BaseConnection(Generic[Row]):
         self._set_isolation_level(value)
 
     def _set_isolation_level(self, value: Optional[IsolationLevel]) -> None:
-        # Base implementation, not thread safe.
-        # Subclasses must call it holding a lock
-        self._check_intrans("isolation_level")
+        raise NotImplementedError
+
+    def _set_isolation_level_gen(self, value: Optional[IsolationLevel]) -> PQGen[None]:
+        yield from self._check_intrans_gen("isolation_level")
         self._isolation_level = IsolationLevel(value) if value is not None else None
         self._begin_statement = b""
 
@@ -201,10 +229,11 @@ class BaseConnection(Generic[Row]):
         self._set_read_only(value)
 
     def _set_read_only(self, value: Optional[bool]) -> None:
-        # Base implementation, not thread safe.
-        # Subclasses must call it holding a lock
-        self._check_intrans("read_only")
-        self._read_only = bool(value)
+        raise NotImplementedError
+
+    def _set_read_only_gen(self, value: Optional[bool]) -> PQGen[None]:
+        yield from self._check_intrans_gen("read_only")
+        self._read_only = bool(value) if value is not None else None
         self._begin_statement = b""
 
     @property
@@ -219,16 +248,20 @@ class BaseConnection(Generic[Row]):
         self._set_deferrable(value)
 
     def _set_deferrable(self, value: Optional[bool]) -> None:
-        # Base implementation, not thread safe.
-        # Subclasses must call it holding a lock
-        self._check_intrans("deferrable")
-        self._deferrable = bool(value)
+        raise NotImplementedError
+
+    def _set_deferrable_gen(self, value: Optional[bool]) -> PQGen[None]:
+        yield from self._check_intrans_gen("deferrable")
+        self._deferrable = bool(value) if value is not None else None
         self._begin_statement = b""
 
-    def _check_intrans(self, attribute: str) -> None:
+    def _check_intrans_gen(self, attribute: str) -> PQGen[None]:
         # Raise an exception if we are in a transaction
         status = self.pgconn.transaction_status
-        if status != TransactionStatus.IDLE:
+        if status == IDLE and self._pipeline:
+            yield from self._pipeline._sync_gen()
+            status = self.pgconn.transaction_status
+        if status != IDLE:
             if self._num_transactions:
                 raise e.ProgrammingError(
                     f"can't change {attribute!r} now: "
@@ -238,7 +271,7 @@ class BaseConnection(Generic[Row]):
                 raise e.ProgrammingError(
                     f"can't change {attribute!r} now: "
                     "connection in transaction status "
-                    f"{TransactionStatus(status).name}"
+                    f"{pq.TransactionStatus(status).name}"
                 )
 
     @property
@@ -248,6 +281,9 @@ class BaseConnection(Generic[Row]):
 
     @property
     def adapters(self) -> AdaptersMap:
+        if not self._adapters:
+            self._adapters = AdaptersMap(postgres.adapters)
+
         return self._adapters
 
     @property
@@ -272,8 +308,22 @@ class BaseConnection(Generic[Row]):
         if self.closed:
             return
 
-        c = self.pgconn.get_cancel()
-        c.cancel()
+        if self._tpc and self._tpc[1]:
+            raise e.ProgrammingError(
+                "cancel() cannot be used with a prepared two-phase transaction"
+            )
+
+        self._try_cancel(self.pgconn)
+
+    @classmethod
+    def _try_cancel(cls, pgconn: "PGconn") -> None:
+        try:
+            # Can fail if the connection is closed
+            c = pgconn.get_cancel()
+        except Exception as ex:
+            logger.warning("couldn't try to cancel query: %s", ex)
+        else:
+            c.cancel()
 
     def add_notice_handler(self, callback: NoticeHandler) -> None:
         """
@@ -298,7 +348,7 @@ class BaseConnection(Generic[Row]):
         wself: "ReferenceType[BaseConnection[Row]]", res: "PGresult"
     ) -> None:
         self = wself()
-        if not (self and self._notice_handler):
+        if not (self and self._notice_handlers):
             return
 
         diag = e.Diagnostic(res, pgconn_encoding(self.pgconn))
@@ -377,25 +427,19 @@ class BaseConnection(Generic[Row]):
     # to the connections subclass, which might wait either in blocking mode
     # or through asyncio.
     #
-    # All these generators assume exclusive acces to the connection: subclasses
+    # All these generators assume exclusive access to the connection: subclasses
     # should have a lock and hold it before calling and consuming them.
 
     @classmethod
-    def _connect_gen(
-        cls: Type[ConnectionType],
-        conninfo: str = "",
-        *,
-        autocommit: bool = False,
-    ) -> PQGenConn[ConnectionType]:
+    def _connect_gen(cls, conninfo: str = "") -> PQGenConn[Self]:
         """Generator to connect to the database and create a new instance."""
         pgconn = yield from connect(conninfo)
         conn = cls(pgconn)
-        conn._autocommit = bool(autocommit)
         return conn
 
     def _exec_command(
-        self, command: Query, result_format: Format = Format.TEXT
-    ) -> PQGen["PGresult"]:
+        self, command: Query, result_format: pq.Format = TEXT
+    ) -> PQGen[Optional["PGresult"]]:
         """
         Generator to send a command and receive the result to the backend.
 
@@ -409,27 +453,35 @@ class BaseConnection(Generic[Row]):
         elif isinstance(command, Composable):
             command = command.as_bytes(self)
 
-        if result_format == Format.TEXT:
-            self.pgconn.send_query(command)
-        else:
-            self.pgconn.send_query_params(command, None, result_format=result_format)
+        if self._pipeline:
+            cmd = partial(
+                self.pgconn.send_query_params,
+                command,
+                None,
+                result_format=result_format,
+            )
+            self._pipeline.command_queue.append(cmd)
+            self._pipeline.result_queue.append(None)
+            return None
+
+        self.pgconn.send_query_params(command, None, result_format=result_format)
 
         result = (yield from execute(self.pgconn))[-1]
-        if result.status not in (ExecStatus.COMMAND_OK, ExecStatus.TUPLES_OK):
-            if result.status == ExecStatus.FATAL_ERROR:
+        if result.status != COMMAND_OK and result.status != TUPLES_OK:
+            if result.status == FATAL_ERROR:
                 raise e.error_from_result(result, encoding=pgconn_encoding(self.pgconn))
             else:
                 raise e.InterfaceError(
-                    f"unexpected result {ExecStatus(result.status).name}"
+                    f"unexpected result {pq.ExecStatus(result.status).name}"
                     f" from command {command.decode()!r}"
                 )
         return result
 
     def _check_connection_ok(self) -> None:
-        if self.pgconn.status == ConnStatus.OK:
+        if self.pgconn.status == OK:
             return
 
-        if self.pgconn.status == ConnStatus.BAD:
+        if self.pgconn.status == BAD:
             raise e.OperationalError("the connection is closed")
         raise e.InterfaceError(
             "cannot execute operations: the connection is"
@@ -441,10 +493,12 @@ class BaseConnection(Generic[Row]):
         if self._autocommit:
             return
 
-        if self.pgconn.transaction_status != TransactionStatus.IDLE:
+        if self.pgconn.transaction_status != IDLE:
             return
 
         yield from self._exec_command(self._get_tx_start_command())
+        if self._pipeline:
+            yield from self._pipeline._sync_gen()
 
     def _get_tx_start_command(self) -> bytes:
         if self._begin_statement:
@@ -474,11 +528,18 @@ class BaseConnection(Generic[Row]):
                 "context. (Transaction will be automatically committed "
                 "on successful exit from context.)"
             )
-        if self.pgconn.transaction_status == TransactionStatus.IDLE:
+        if self._tpc:
+            raise e.ProgrammingError(
+                "commit() cannot be used during a two-phase transaction"
+            )
+        if self.pgconn.transaction_status == IDLE:
             return
 
         yield from self._exec_command(b"COMMIT")
 
+        if self._pipeline:
+            yield from self._pipeline._sync_gen()
+
     def _rollback_gen(self) -> PQGen[None]:
         """Generator implementing `Connection.rollback()`."""
         if self._num_transactions:
@@ -487,14 +548,110 @@ class BaseConnection(Generic[Row]):
                 "context. (Either raise Rollback() or allow "
                 "an exception to propagate out of the context.)"
             )
-        if self.pgconn.transaction_status == TransactionStatus.IDLE:
+        if self._tpc:
+            raise e.ProgrammingError(
+                "rollback() cannot be used during a two-phase transaction"
+            )
+
+        # Get out of a "pipeline aborted" state
+        if self._pipeline:
+            yield from self._pipeline._sync_gen()
+
+        if self.pgconn.transaction_status == IDLE:
             return
 
         yield from self._exec_command(b"ROLLBACK")
-        cmd = self._prepared.clear()
-        if cmd:
+        self._prepared.clear()
+        for cmd in self._prepared.get_maintenance_commands():
             yield from self._exec_command(cmd)
 
+        if self._pipeline:
+            yield from self._pipeline._sync_gen()
+
+    def xid(self, format_id: int, gtrid: str, bqual: str) -> Xid:
+        """
+        Returns a `Xid` to pass to the `!tpc_*()` methods of this connection.
+
+        The argument types and constraints are explained in
+        :ref:`two-phase-commit`.
+
+        The values passed to the method will be available on the returned
+        object as the members `~Xid.format_id`, `~Xid.gtrid`, `~Xid.bqual`.
+        """
+        self._check_tpc()
+        return Xid.from_parts(format_id, gtrid, bqual)
+
+    def _tpc_begin_gen(self, xid: Union[Xid, str]) -> PQGen[None]:
+        self._check_tpc()
+
+        if not isinstance(xid, Xid):
+            xid = Xid.from_string(xid)
+
+        if self.pgconn.transaction_status != IDLE:
+            raise e.ProgrammingError(
+                "can't start two-phase transaction: connection in status"
+                f" {pq.TransactionStatus(self.pgconn.transaction_status).name}"
+            )
+
+        if self._autocommit:
+            raise e.ProgrammingError(
+                "can't use two-phase transactions in autocommit mode"
+            )
+
+        self._tpc = (xid, False)
+        yield from self._exec_command(self._get_tx_start_command())
+
+    def _tpc_prepare_gen(self) -> PQGen[None]:
+        if not self._tpc:
+            raise e.ProgrammingError(
+                "'tpc_prepare()' must be called inside a two-phase transaction"
+            )
+        if self._tpc[1]:
+            raise e.ProgrammingError(
+                "'tpc_prepare()' cannot be used during a prepared two-phase transaction"
+            )
+        xid = self._tpc[0]
+        self._tpc = (xid, True)
+        yield from self._exec_command(SQL("PREPARE TRANSACTION {}").format(str(xid)))
+        if self._pipeline:
+            yield from self._pipeline._sync_gen()
+
+    def _tpc_finish_gen(
+        self, action: LiteralString, xid: Union[Xid, str, None]
+    ) -> PQGen[None]:
+        fname = f"tpc_{action.lower()}()"
+        if xid is None:
+            if not self._tpc:
+                raise e.ProgrammingError(
+                    f"{fname} without xid must must be"
+                    " called inside a two-phase transaction"
+                )
+            xid = self._tpc[0]
+        else:
+            if self._tpc:
+                raise e.ProgrammingError(
+                    f"{fname} with xid must must be called"
+                    " outside a two-phase transaction"
+                )
+            if not isinstance(xid, Xid):
+                xid = Xid.from_string(xid)
+
+        if self._tpc and not self._tpc[1]:
+            meth: Callable[[], PQGen[None]]
+            meth = getattr(self, f"_{action.lower()}_gen")
+            self._tpc = None
+            yield from meth()
+        else:
+            yield from self._exec_command(
+                SQL("{} PREPARED {}").format(SQL(action), str(xid))
+            )
+            self._tpc = None
+
+    def _check_tpc(self) -> None:
+        """Raise NotSupportedError if TPC is not supported."""
+        # TPC supported on every supported PostgreSQL version.
+        pass
+
 
 class Connection(BaseConnection[Row]):
     """
@@ -506,10 +663,15 @@ class Connection(BaseConnection[Row]):
     cursor_factory: Type[Cursor[Row]]
     server_cursor_factory: Type[ServerCursor[Row]]
     row_factory: RowFactory[Row]
+    _pipeline: Optional[Pipeline]
 
-    def __init__(self, pgconn: "PGconn", row_factory: Optional[RowFactory[Row]] = None):
+    def __init__(
+        self,
+        pgconn: "PGconn",
+        row_factory: RowFactory[Row] = cast(RowFactory[Row], tuple_row),
+    ):
         super().__init__(pgconn)
-        self.row_factory = row_factory or cast(RowFactory[Row], tuple_row)
+        self.row_factory = row_factory
         self.lock = threading.Lock()
         self.cursor_factory = Cursor
         self.server_cursor_factory = ServerCursor
@@ -522,9 +684,14 @@ class Connection(BaseConnection[Row]):
         *,
         autocommit: bool = False,
         row_factory: RowFactory[Row],
+        prepare_threshold: Optional[int] = 5,
+        cursor_factory: Optional[Type[Cursor[Row]]] = None,
         context: Optional[AdaptContext] = None,
         **kwargs: Union[None, int, str],
     ) -> "Connection[Row]":
+        # TODO: returned type should be Self. See #308.
+        # Unfortunately we cannot use Self[Row] as Self is not parametric.
+        # https://peps.python.org/pep-0673/#use-in-generic-classes
         ...
 
     @overload
@@ -534,10 +701,11 @@ class Connection(BaseConnection[Row]):
         conninfo: str = "",
         *,
         autocommit: bool = False,
+        prepare_threshold: Optional[int] = 5,
+        cursor_factory: Optional[Type[Cursor[Any]]] = None,
         context: Optional[AdaptContext] = None,
         **kwargs: Union[None, int, str],
-    ) -> "Connection[TupleRow]":
-        ...
+    ) -> "Connection[TupleRow]": ...
 
     @classmethod  # type: ignore[misc] # https://github.com/python/mypy/issues/11004
     def connect(
@@ -545,31 +713,51 @@ class Connection(BaseConnection[Row]):
         conninfo: str = "",
         *,
         autocommit: bool = False,
+        prepare_threshold: Optional[int] = 5,
         row_factory: Optional[RowFactory[Row]] = None,
+        cursor_factory: Optional[Type[Cursor[Row]]] = None,
         context: Optional[AdaptContext] = None,
         **kwargs: Any,
-    ) -> "Connection[Any]":
+    ) -> Self:
         """
         Connect to a database server and return a new `Connection` instance.
         """
         params = cls._get_connection_params(conninfo, **kwargs)
-        conninfo = make_conninfo(**params)
+        timeout = timeout_from_conninfo(params)
+        rv = None
+        attempts = conninfo_attempts(params)
+        for attempt in attempts:
+            try:
+                conninfo = make_conninfo(**attempt)
+                rv = cls._wait_conn(cls._connect_gen(conninfo), timeout=timeout)
+                break
+            except e._NO_TRACEBACK as ex:
+                if len(attempts) > 1:
+                    logger.debug(
+                        "connection attempt failed on host: %r, port: %r,"
+                        " hostaddr: %r: %s",
+                        attempt.get("host"),
+                        attempt.get("port"),
+                        attempt.get("hostaddr"),
+                        str(ex),
+                    )
+                last_ex = ex
 
-        try:
-            rv = cls._wait_conn(
-                cls._connect_gen(conninfo, autocommit=autocommit),
-                timeout=params["connect_timeout"],
-            )
-        except e.Error as ex:
-            raise ex.with_traceback(None)
+        if not rv:
+            assert last_ex
+            raise last_ex.with_traceback(None)
 
+        rv._autocommit = bool(autocommit)
         if row_factory:
             rv.row_factory = row_factory
+        if cursor_factory:
+            rv.cursor_factory = cursor_factory
         if context:
             rv._adapters = AdaptersMap(context.adapters)
+        rv.prepare_threshold = prepare_threshold
         return rv
 
-    def __enter__(self) -> "Connection[Row]":
+    def __enter__(self) -> Self:
         return self
 
     def __exit__(
@@ -600,7 +788,7 @@ class Connection(BaseConnection[Row]):
             self.close()
 
     @classmethod
-    def _get_connection_params(cls, conninfo: str, **kwargs: Any) -> Dict[str, Any]:
+    def _get_connection_params(cls, conninfo: str, **kwargs: Any) -> ConnDict:
         """Manipulate connection parameters before connecting.
 
         :param conninfo: Connection string as received by `~Connection.connect()`.
@@ -608,32 +796,25 @@ class Connection(BaseConnection[Row]):
         :return: Connection arguments merged and eventually modified, in a
             format similar to `~conninfo.conninfo_to_dict()`.
         """
-        params = conninfo_to_dict(conninfo, **kwargs)
-
-        # Make sure there is an usable connect_timeout
-        if "connect_timeout" in params:
-            params["connect_timeout"] = int(params["connect_timeout"])
-        else:
-            params["connect_timeout"] = None
-
-        return params
+        return conninfo_to_dict(conninfo, **kwargs)
 
     def close(self) -> None:
         """Close the database connection."""
         if self.closed:
             return
         self._closed = True
+
+        # TODO: maybe send a cancel on close, if the connection is ACTIVE?
+
         self.pgconn.finish()
 
     @overload
-    def cursor(self, *, binary: bool = False) -> Cursor[Row]:
-        ...
+    def cursor(self, *, binary: bool = False) -> Cursor[Row]: ...
 
     @overload
     def cursor(
         self, *, binary: bool = False, row_factory: RowFactory[CursorRow]
-    ) -> Cursor[CursorRow]:
-        ...
+    ) -> Cursor[CursorRow]: ...
 
     @overload
     def cursor(
@@ -643,8 +824,7 @@ class Connection(BaseConnection[Row]):
         binary: bool = False,
         scrollable: Optional[bool] = None,
         withhold: bool = False,
-    ) -> ServerCursor[Row]:
-        ...
+    ) -> ServerCursor[Row]: ...
 
     @overload
     def cursor(
@@ -655,8 +835,7 @@ class Connection(BaseConnection[Row]):
         row_factory: RowFactory[CursorRow],
         scrollable: Optional[bool] = None,
         withhold: bool = False,
-    ) -> ServerCursor[CursorRow]:
-        ...
+    ) -> ServerCursor[CursorRow]: ...
 
     def cursor(
         self,
@@ -688,7 +867,7 @@ class Connection(BaseConnection[Row]):
             cur = self.cursor_factory(self, row_factory=row_factory)
 
         if binary:
-            cur.format = Format.BINARY
+            cur.format = BINARY
 
         return cur
 
@@ -704,11 +883,11 @@ class Connection(BaseConnection[Row]):
         try:
             cur = self.cursor()
             if binary:
-                cur.format = Format.BINARY
+                cur.format = BINARY
 
             return cur.execute(query, params, prepare=prepare)
 
-        except e.Error as ex:
+        except e._NO_TRACEBACK as ex:
             raise ex.with_traceback(None)
 
     def commit(self) -> None:
@@ -736,24 +915,49 @@ class Connection(BaseConnection[Row]):
             block even if there were no error (e.g. to try a no-op process).
         :rtype: Transaction
         """
-        with Transaction(self, savepoint_name, force_rollback) as tx:
-            yield tx
+        tx = Transaction(self, savepoint_name, force_rollback)
+        if self._pipeline:
+            with self.pipeline(), tx, self.pipeline():
+                yield tx
+        else:
+            with tx:
+                yield tx
 
     def notifies(self) -> Generator[Notify, None, None]:
         """
         Yield `Notify` objects as soon as they are received from the database.
         """
-        while 1:
+        while True:
             with self.lock:
                 try:
                     ns = self.wait(notifies(self.pgconn))
-                except e.Error as ex:
+                except e._NO_TRACEBACK as ex:
                     raise ex.with_traceback(None)
             enc = pgconn_encoding(self.pgconn)
             for pgn in ns:
                 n = Notify(pgn.relname.decode(enc), pgn.extra.decode(enc), pgn.be_pid)
                 yield n
 
+    @contextmanager
+    def pipeline(self) -> Iterator[Pipeline]:
+        """Switch the connection into pipeline mode."""
+        with self.lock:
+            self._check_connection_ok()
+
+            pipeline = self._pipeline
+            if pipeline is None:
+                # WARNING: reference loop, broken ahead.
+                pipeline = self._pipeline = Pipeline(self)
+
+        try:
+            with pipeline:
+                yield pipeline
+        finally:
+            if pipeline.level == 0:
+                with self.lock:
+                    assert pipeline is self._pipeline
+                    self._pipeline = None
+
     def wait(self, gen: PQGen[RV], timeout: Optional[float] = 0.1) -> RV:
         """
         Consume a generator operating on the connection.
@@ -765,9 +969,8 @@ class Connection(BaseConnection[Row]):
             return waiting.wait(gen, self.pgconn.socket, timeout=timeout)
         except KeyboardInterrupt:
             # On Ctrl-C, try to cancel the query in the server, otherwise
-            # otherwise the connection will be stuck in ACTIVE state
-            c = self.pgconn.get_cancel()
-            c.cancel()
+            # the connection will remain stuck in ACTIVE state.
+            self._try_cancel(self.pgconn)
             try:
                 waiting.wait(gen, self.pgconn.socket, timeout=timeout)
             except e.QueryCanceled:
@@ -781,16 +984,59 @@ class Connection(BaseConnection[Row]):
 
     def _set_autocommit(self, value: bool) -> None:
         with self.lock:
-            super()._set_autocommit(value)
+            self.wait(self._set_autocommit_gen(value))
 
     def _set_isolation_level(self, value: Optional[IsolationLevel]) -> None:
         with self.lock:
-            super()._set_isolation_level(value)
+            self.wait(self._set_isolation_level_gen(value))
 
     def _set_read_only(self, value: Optional[bool]) -> None:
         with self.lock:
-            super()._set_read_only(value)
+            self.wait(self._set_read_only_gen(value))
 
     def _set_deferrable(self, value: Optional[bool]) -> None:
         with self.lock:
-            super()._set_deferrable(value)
+            self.wait(self._set_deferrable_gen(value))
+
+    def tpc_begin(self, xid: Union[Xid, str]) -> None:
+        """
+        Begin a TPC transaction with the given transaction ID `!xid`.
+        """
+        with self.lock:
+            self.wait(self._tpc_begin_gen(xid))
+
+    def tpc_prepare(self) -> None:
+        """
+        Perform the first phase of a transaction started with `tpc_begin()`.
+        """
+        try:
+            with self.lock:
+                self.wait(self._tpc_prepare_gen())
+        except e.ObjectNotInPrerequisiteState as ex:
+            raise e.NotSupportedError(str(ex)) from None
+
+    def tpc_commit(self, xid: Union[Xid, str, None] = None) -> None:
+        """
+        Commit a prepared two-phase transaction.
+        """
+        with self.lock:
+            self.wait(self._tpc_finish_gen("COMMIT", xid))
+
+    def tpc_rollback(self, xid: Union[Xid, str, None] = None) -> None:
+        """
+        Roll back a prepared two-phase transaction.
+        """
+        with self.lock:
+            self.wait(self._tpc_finish_gen("ROLLBACK", xid))
+
+    def tpc_recover(self) -> List[Xid]:
+        self._check_tpc()
+        status = self.info.transaction_status
+        with self.cursor(row_factory=args_row(Xid._from_record)) as cur:
+            cur.execute(Xid._get_recover_query())
+            res = cur.fetchall()
+
+        if status == IDLE and self.info.transaction_status == INTRANS:
+            self.rollback()
+
+        return res
