@@ -6,7 +6,7 @@ psycopg async cursor objects
 
 from types import TracebackType
 from typing import Any, AsyncIterator, Iterable, List
-from typing import Optional, Type, TypeVar, TYPE_CHECKING, overload
+from typing import Optional, Type, TYPE_CHECKING, overload
 from contextlib import asynccontextmanager
 
 from . import pq
@@ -15,6 +15,7 @@ from .abc import Query, Params
 from .copy import AsyncCopy, AsyncWriter as AsyncCopyWriter
 from .rows import Row, RowMaker, AsyncRowFactory
 from .cursor import BaseCursor
+from ._compat import Self
 from ._pipeline import Pipeline
 
 if TYPE_CHECKING:
@@ -26,20 +27,14 @@ ACTIVE = pq.TransactionStatus.ACTIVE
 class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
     __module__ = "psycopg"
     __slots__ = ()
-    _Self = TypeVar("_Self", bound="AsyncCursor[Any]")
 
     @overload
-    def __init__(self: "AsyncCursor[Row]", connection: "AsyncConnection[Row]"):
-        ...
+    def __init__(self, connection: "AsyncConnection[Row]"): ...
 
     @overload
     def __init__(
-        self: "AsyncCursor[Row]",
-        connection: "AsyncConnection[Any]",
-        *,
-        row_factory: AsyncRowFactory[Row],
-    ):
-        ...
+        self, connection: "AsyncConnection[Any]", *, row_factory: AsyncRowFactory[Row]
+    ): ...
 
     def __init__(
         self,
@@ -50,7 +45,7 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         super().__init__(connection)
         self._row_factory = row_factory or connection.row_factory
 
-    async def __aenter__(self: _Self) -> _Self:
+    async def __aenter__(self) -> Self:
         return self
 
     async def __aexit__(
@@ -78,19 +73,19 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         return self._row_factory(self)
 
     async def execute(
-        self: _Self,
+        self,
         query: Query,
         params: Optional[Params] = None,
         *,
         prepare: Optional[bool] = None,
         binary: Optional[bool] = None,
-    ) -> _Self:
+    ) -> Self:
         try:
             async with self._conn.lock:
                 await self._conn.wait(
                     self._execute_gen(query, params, prepare=prepare, binary=binary)
                 )
-        except e.Error as ex:
+        except e._NO_TRACEBACK as ex:
             raise ex.with_traceback(None)
         return self
 
@@ -118,10 +113,11 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
                             self._executemany_gen_pipeline(query, params_seq, returning)
                         )
             else:
-                await self._conn.wait(
-                    self._executemany_gen_no_pipeline(query, params_seq, returning)
-                )
-        except e.Error as ex:
+                async with self._conn.lock:
+                    await self._conn.wait(
+                        self._executemany_gen_no_pipeline(query, params_seq, returning)
+                    )
+        except e._NO_TRACEBACK as ex:
             raise ex.with_traceback(None)
 
     async def stream(
@@ -134,8 +130,8 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
         if self._pgconn.pipeline_status:
             raise e.ProgrammingError("stream() cannot be used in pipeline mode")
 
-        try:
-            async with self._conn.lock:
+        async with self._conn.lock:
+            try:
                 await self._conn.wait(
                     self._stream_send_gen(query, params, binary=binary)
                 )
@@ -146,33 +142,36 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
                     yield rec
                     first = False
 
-        except e.Error as ex:
-            raise ex.with_traceback(None)
+            except e._NO_TRACEBACK as ex:
+                raise ex.with_traceback(None)
 
-        finally:
-            if self._pgconn.transaction_status == ACTIVE:
-                # Try to cancel the query, then consume the results already received.
-                self._conn.cancel()
-                try:
-                    while await self._conn.wait(self._stream_fetchone_gen(first=False)):
+            finally:
+                if self._pgconn.transaction_status == ACTIVE:
+                    # Try to cancel the query, then consume the results
+                    # already received.
+                    self._conn.cancel()
+                    try:
+                        while await self._conn.wait(
+                            self._stream_fetchone_gen(first=False)
+                        ):
+                            pass
+                    except Exception:
                         pass
-                except Exception:
-                    pass
 
-                # Try to get out of ACTIVE state. Just do a single attempt, which
-                # should work to recover from an error or query cancelled.
-                try:
-                    await self._conn.wait(self._stream_fetchone_gen(first=False))
-                except Exception:
-                    pass
+                    # Try to get out of ACTIVE state. Just do a single attempt, which
+                    # should work to recover from an error or query cancelled.
+                    try:
+                        await self._conn.wait(self._stream_fetchone_gen(first=False))
+                    except Exception:
+                        pass
 
     async def fetchone(self) -> Optional[Row]:
         await self._fetch_pipeline()
         self._check_result_for_fetch()
-        rv = self._tx.load_row(self._pos, self._make_row)
-        if rv is not None:
+        record = self._tx.load_row(self._pos, self._make_row)
+        if record is not None:
             self._pos += 1
-        return rv
+        return record
 
     async def fetchmany(self, size: int = 0) -> List[Row]:
         await self._fetch_pipeline()
@@ -212,6 +211,7 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
             yield row
 
     async def scroll(self, value: int, mode: str = "relative") -> None:
+        await self._fetch_pipeline()
         self._scroll(value, mode)
 
     @asynccontextmanager
@@ -231,7 +231,7 @@ class AsyncCursor(BaseCursor["AsyncConnection[Any]", Row]):
 
             async with AsyncCopy(self, writer=writer) as copy:
                 yield copy
-        except e.Error as ex:
+        except e._NO_TRACEBACK as ex:
             raise ex.with_traceback(None)
 
         self._select_current_result(0)
