@@ -22,12 +22,12 @@ if TYPE_CHECKING:
 
 _struct_timetz = struct.Struct("!qi")  # microseconds, sec tz offset
 _pack_timetz = cast(Callable[[int, int], bytes], _struct_timetz.pack)
-_unpack_timetz = cast(Callable[[bytes], Tuple[int, int]], _struct_timetz.unpack)
+_unpack_timetz = cast(Callable[[Buffer], Tuple[int, int]], _struct_timetz.unpack)
 
 _struct_interval = struct.Struct("!qii")  # microseconds, days, months
 _pack_interval = cast(Callable[[int, int, int], bytes], _struct_interval.pack)
 _unpack_interval = cast(
-    Callable[[bytes], Tuple[int, int, int]], _struct_interval.unpack
+    Callable[[Buffer], Tuple[int, int, int]], _struct_interval.unpack
 )
 
 utc = timezone.utc
@@ -38,7 +38,6 @@ _py_date_min_days = date.min.toordinal()
 
 
 class DateDumper(Dumper):
-
     oid = postgres.types["date"].oid
 
     def dump(self, obj: date) -> bytes:
@@ -48,7 +47,6 @@ class DateDumper(Dumper):
 
 
 class DateBinaryDumper(Dumper):
-
     format = Format.BINARY
     oid = postgres.types["date"].oid
 
@@ -69,6 +67,14 @@ class _BaseTimeDumper(Dumper):
     def upgrade(self, obj: time, format: PyFormat) -> Dumper:
         raise NotImplementedError
 
+    def _get_offset(self, obj: time) -> timedelta:
+        offset = obj.utcoffset()
+        if offset is None:
+            raise DataError(
+                f"cannot calculate the offset of tzinfo '{obj.tzinfo}' without a date"
+            )
+        return offset
+
 
 class _BaseTimeTextDumper(_BaseTimeDumper):
     def dump(self, obj: time) -> bytes:
@@ -76,7 +82,6 @@ class _BaseTimeTextDumper(_BaseTimeDumper):
 
 
 class TimeDumper(_BaseTimeTextDumper):
-
     oid = postgres.types["time"].oid
 
     def upgrade(self, obj: time, format: PyFormat) -> Dumper:
@@ -87,12 +92,14 @@ class TimeDumper(_BaseTimeTextDumper):
 
 
 class TimeTzDumper(_BaseTimeTextDumper):
-
     oid = postgres.types["timetz"].oid
 
+    def dump(self, obj: time) -> bytes:
+        self._get_offset(obj)
+        return super().dump(obj)
+
 
 class TimeBinaryDumper(_BaseTimeDumper):
-
     format = Format.BINARY
     oid = postgres.types["time"].oid
 
@@ -110,7 +117,6 @@ class TimeBinaryDumper(_BaseTimeDumper):
 
 
 class TimeTzBinaryDumper(_BaseTimeDumper):
-
     format = Format.BINARY
     oid = postgres.types["timetz"].oid
 
@@ -118,8 +124,7 @@ class TimeTzBinaryDumper(_BaseTimeDumper):
         us = obj.microsecond + 1_000_000 * (
             obj.second + 60 * (obj.minute + 60 * obj.hour)
         )
-        off = obj.utcoffset()
-        assert off is not None
+        off = self._get_offset(obj)
         return _pack_timetz(us, -int(off.total_seconds()))
 
 
@@ -144,7 +149,6 @@ class _BaseDatetimeTextDumper(_BaseDatetimeDumper):
 
 
 class DatetimeDumper(_BaseDatetimeTextDumper):
-
     oid = postgres.types["timestamptz"].oid
 
     def upgrade(self, obj: datetime, format: PyFormat) -> Dumper:
@@ -155,12 +159,10 @@ class DatetimeDumper(_BaseDatetimeTextDumper):
 
 
 class DatetimeNoTzDumper(_BaseDatetimeTextDumper):
-
     oid = postgres.types["timestamp"].oid
 
 
 class DatetimeBinaryDumper(_BaseDatetimeDumper):
-
     format = Format.BINARY
     oid = postgres.types["timestamptz"].oid
 
@@ -177,7 +179,6 @@ class DatetimeBinaryDumper(_BaseDatetimeDumper):
 
 
 class DatetimeNoTzBinaryDumper(_BaseDatetimeDumper):
-
     format = Format.BINARY
     oid = postgres.types["timestamp"].oid
 
@@ -188,24 +189,26 @@ class DatetimeNoTzBinaryDumper(_BaseDatetimeDumper):
 
 
 class TimedeltaDumper(Dumper):
-
     oid = postgres.types["interval"].oid
 
     def __init__(self, cls: type, context: Optional[AdaptContext] = None):
         super().__init__(cls, context)
-        if self.connection:
-            if (
-                self.connection.pgconn.parameter_status(b"IntervalStyle")
-                == b"sql_standard"
-            ):
-                setattr(self, "dump", self._dump_sql)
+        if _get_intervalstyle(self.connection) == b"sql_standard":
+            self._dump_method = self._dump_sql
+        else:
+            self._dump_method = self._dump_any
 
     def dump(self, obj: timedelta) -> bytes:
+        return self._dump_method(self, obj)
+
+    @staticmethod
+    def _dump_any(self: "TimedeltaDumper", obj: timedelta) -> bytes:
         # The comma is parsed ok by PostgreSQL but it's not documented
         # and it seems brittle to rely on it. CRDB doesn't consume it well.
         return str(obj).encode().replace(b",", b"")
 
-    def _dump_sql(self, obj: timedelta) -> bytes:
+    @staticmethod
+    def _dump_sql(self: "TimedeltaDumper", obj: timedelta) -> bytes:
         # sql_standard format needs explicit signs
         # otherwise -1 day 1 sec will mean -1 sec
         return b"%+d day %+d second %+d microsecond" % (
@@ -216,7 +219,6 @@ class TimedeltaDumper(Dumper):
 
 
 class TimedeltaBinaryDumper(Dumper):
-
     format = Format.BINARY
     oid = postgres.types["interval"].oid
 
@@ -226,7 +228,6 @@ class TimedeltaBinaryDumper(Dumper):
 
 
 class DateLoader(Loader):
-
     _ORDER_YMD = 0
     _ORDER_DMY = 1
     _ORDER_MDY = 2
@@ -270,7 +271,6 @@ class DateLoader(Loader):
 
 
 class DateBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> date:
@@ -285,7 +285,6 @@ class DateBinaryLoader(Loader):
 
 
 class TimeLoader(Loader):
-
     _re_format = re.compile(rb"^(\d+):(\d+):(\d+)(?:\.(\d+))?")
 
     def load(self, data: Buffer) -> time:
@@ -312,7 +311,6 @@ class TimeLoader(Loader):
 
 
 class TimeBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> time:
@@ -327,7 +325,6 @@ class TimeBinaryLoader(Loader):
 
 
 class TimetzLoader(Loader):
-
     _re_format = re.compile(
         rb"""(?ix)
         ^
@@ -369,7 +366,6 @@ class TimetzLoader(Loader):
 
 
 class TimetzBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> time:
@@ -386,7 +382,6 @@ class TimetzBinaryLoader(Loader):
 
 
 class TimestampLoader(Loader):
-
     _re_format = re.compile(
         rb"""(?ix)
         ^
@@ -472,7 +467,6 @@ class TimestampLoader(Loader):
 
 
 class TimestampBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> datetime:
@@ -487,7 +481,6 @@ class TimestampBinaryLoader(Loader):
 
 
 class TimestamptzLoader(Loader):
-
     _re_format = re.compile(
         rb"""(?ix)
         ^
@@ -505,10 +498,16 @@ class TimestamptzLoader(Loader):
         self._timezone = get_tzinfo(self.connection.pgconn if self.connection else None)
 
         ds = _get_datestyle(self.connection)
-        if not ds.startswith(b"I"):  # not ISO
-            setattr(self, "load", self._load_notimpl)
+        if ds.startswith(b"I"):  # ISO
+            self._load_method = self._load_iso
+        else:
+            self._load_method = self._load_notimpl
 
     def load(self, data: Buffer) -> datetime:
+        return self._load_method(self, data)
+
+    @staticmethod
+    def _load_iso(self: "TimestamptzLoader", data: Buffer) -> datetime:
         m = self._re_format.match(data)
         if not m:
             raise _get_timestamp_load_error(self.connection, data) from None
@@ -554,7 +553,8 @@ class TimestamptzLoader(Loader):
 
         raise _get_timestamp_load_error(self.connection, data, ex) from None
 
-    def _load_notimpl(self, data: Buffer) -> datetime:
+    @staticmethod
+    def _load_notimpl(self: "TimestamptzLoader", data: Buffer) -> datetime:
         s = bytes(data).decode("utf8", "replace")
         ds = _get_datestyle(self.connection).decode("ascii")
         raise NotImplementedError(
@@ -563,7 +563,6 @@ class TimestamptzLoader(Loader):
 
 
 class TimestamptzBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def __init__(self, oid: int, context: Optional[AdaptContext] = None):
@@ -600,7 +599,6 @@ class TimestamptzBinaryLoader(Loader):
 
 
 class IntervalLoader(Loader):
-
     _re_interval = re.compile(
         rb"""
         (?: ([-+]?\d+) \s+ years? \s* )?                # Years
@@ -614,12 +612,16 @@ class IntervalLoader(Loader):
 
     def __init__(self, oid: int, context: Optional[AdaptContext] = None):
         super().__init__(oid, context)
-        if self.connection:
-            ints = self.connection.pgconn.parameter_status(b"IntervalStyle")
-            if ints != b"postgres":
-                setattr(self, "load", self._load_notimpl)
+        if _get_intervalstyle(self.connection) == b"postgres":
+            self._load_method = self._load_postgres
+        else:
+            self._load_method = self._load_notimpl
 
     def load(self, data: Buffer) -> timedelta:
+        return self._load_method(self, data)
+
+    @staticmethod
+    def _load_postgres(self: "IntervalLoader", data: Buffer) -> timedelta:
         m = self._re_interval.match(data)
         if not m:
             s = bytes(data).decode("utf8", "replace")
@@ -647,20 +649,16 @@ class IntervalLoader(Loader):
             s = bytes(data).decode("utf8", "replace")
             raise DataError(f"can't parse interval {s!r}: {e}") from None
 
-    def _load_notimpl(self, data: Buffer) -> timedelta:
+    @staticmethod
+    def _load_notimpl(self: "IntervalLoader", data: Buffer) -> timedelta:
         s = bytes(data).decode("utf8", "replace")
-        ints = (
-            self.connection
-            and self.connection.pgconn.parameter_status(b"IntervalStyle")
-            or b"unknown"
-        ).decode("utf8", "replace")
+        ints = _get_intervalstyle(self.connection).decode("utf8", "replace")
         raise NotImplementedError(
             f"can't parse interval with IntervalStyle {ints}: {s!r}"
         )
 
 
 class IntervalBinaryLoader(Loader):
-
     format = Format.BINARY
 
     def load(self, data: Buffer) -> timedelta:
@@ -687,6 +685,15 @@ def _get_datestyle(conn: Optional["BaseConnection[Any]"]) -> bytes:
     return b"ISO, DMY"
 
 
+def _get_intervalstyle(conn: Optional["BaseConnection[Any]"]) -> bytes:
+    if conn:
+        ints = conn.pgconn.parameter_status(b"IntervalStyle")
+        if ints:
+            return ints
+
+    return b"unknown"
+
+
 def _get_timestamp_load_error(
     conn: Optional["BaseConnection[Any]"], data: Buffer, ex: Optional[Exception] = None
 ) -> Exception:
